#include <iostream>
#include <iomanip>
#include <fstream>
#include <math.h>
#include <stdio.h>
#include <cstdlib>
#include "cell.h"
using namespace std;

int main()
{
	//********************** Information on Hydroshed original data. Should Not Change ******************
	/* HydroSHED data for Asia, Flow Direction and DEM
	LL corner Lat =-10
	LL corner Lon =55
	UR corner Lat =60
	UR corner Lon =180	
	*/
	const int HyRow=8400;    // =((UR_LAT-LL_LAT)*120
	const int HyCol=15000;   // =((UR_LON-LL_LON)*120
	double  Dx = 0.00833333; // cellsize in degrees
	double  ULLat = 60-Dx/2; // The latitude of the center of North West (upper left) grid. 
	                         // 60 (degree) is the latitude of North West corner of Asia data of HydroSHED.
	double  ULLon = 55+Dx/2; // The longitude of the center of North West (upper left) grid. 
	                         // 55 (degree) is the longitude of North West corner of Asia data of HydroSHED.
	const char *projectName = "";   // Used if you creat project folder.
	//***************************************************************************************************

	char filename[256];
	const int NODATA = -9999;

	//read study area for cliping topographic data.
	cout << "1) Reading location of study area." << endl;
	ifstream finClipArea("./clipArea.dat");
	if(!finClipArea){
		cout << "clipArea.dat cannot open." << endl;
		cout << endl;
		cout << "Program ends. Hit any key!" << endl;
		char ch;
		cin.get(ch);
		return 1;
	}

	// define the calculation region //  one grid size is =1/120 degrees.
	///*Chaopria thailand */
	//const int Col = 120*6;	// Number of columns = degrees in Longitude direction * 120 
	//const int Row = 120*8;	// Number of rows = degrees in latitude direction * 120
	//const int SWLON = 97;		// Lower Left corner Longitude (deg.)
	//const int SWLAT = 13;		// Lower Left corner Latitude (deg.)

	int wCol, wRow;
	double wSWLON, wSWLAT;

	finClipArea >> wCol; cout << "  Number of column : " << wCol << endl;
	finClipArea >> wRow; cout << "  Number of row : " << wRow << endl;
	// Lower Left corner longitude (deg.)
	finClipArea >> wSWLON; cout << "  Longitude (degree) of South West corner : " << wSWLON << endl;
	// Lower left corner Latitude (deg.)
	finClipArea >> wSWLAT; cout << "  Latitude (degree) of South West corner : " << wSWLAT << endl;

	const int Col = wCol;     // column number of study area
	const int Row = wRow;     // row number of study area
	const double SWLON = wSWLON; // South west corner of longitude (degree)
	const double SWLAT = wSWLAT; // South west corner of latiitude (degree)

	finClipArea.close();


	/*--------------------------------File Names-------------------------------------------------*/	
	FILE *HyDem;
	FILE *HyDir;
	// errno_t error;

	//make output project directory
	//sprintf(filename,"mkdir ./output/%s",projectName);
	//system (filename);
	
	//memory allocation
	int** HyshedClipDem;
	HyshedClipDem = new int *[Row];
	for(int i=0; i < Row; i++){
		HyshedClipDem[i] = new int [Col];
	}
	int** HyshedClipDir;
	HyshedClipDir = new int *[Row];
	for(int i=0; i < Row; i++){
		HyshedClipDir[i] = new int [Col];
	}

	//read topographic data generated from HydroSHED
	//if(error = fopen_s(&HyDem, "./input/as_dem_30s_clipped_55--10-180-60.bin", "rb") != NULL) {
	if((HyDem = fopen("./input/as_dem_30s_clipped_55--10-180-60.bin", "rb")) == NULL) {
		cout << "Hydroshed DEM cannot open." << endl;  //hunu
		exit(1);
	} 

	//if(error = fopen_s(&HyDir, "./input/as_dir_30s_clipped_55--10-180-60.bin", "rb") != NULL) {
	if((HyDir = fopen("./input/as_dir_30s_clipped_55--10-180-60.bin", "rb")) == NULL) {
		cout << "Hydroshed Flow direction file cannot open." << endl;  //hunu
		exit(1);
	}

	// set clip area	
	short int idem, idir;
	int cut_i, cut_j;   // Upper left (North-West) corner i and  j of the clipped box
	double tempLat, tempLon;
	tempLat = SWLAT + (double) Row/120 - Dx/2;
	tempLon = SWLON + Dx/2;
	
	cut_i= (int) ((ULLat-tempLat)/Dx +0.5);
	cut_j= (int) ((tempLon-ULLon)/Dx +0.5);

	// read the Hydroshed data and assigned to an array
	cout << "2) Reading topography data." << endl;
	for (int i=0; i<HyRow; i++) { // read from north
		if (i%100 == 0) {
			//cout << "  reading " << i << "/" << HyRow<< "th row data.\r";
			cout << "  reading " << i << "/" << HyRow<< "th row data.\n";
		}
		for (int j=0; j<HyCol; j++) { // read from west
			fread(&idem, sizeof(short int), 1, HyDem);	
			fread(&idir, sizeof(short int), 1, HyDir);
			
			if ((i>=cut_i) && (i<cut_i+Row) && (j>=cut_j) && (j<cut_j+Col)) {
				HyshedClipDem[i-cut_i][j-cut_j]=idem;
				HyshedClipDir[i-cut_i][j-cut_j]=idir;
			}
		}
	}	
	fclose(HyDem);
	fclose(HyDir);
	
	//write the clipped data into ESRI ASCII file format
	//sprintf_s(filename, 256, "./output/arcgis/cutDem.asc", projectName);
	sprintf(filename, "./output/arcgis/cutDem.asc");
	ofstream foutDem(filename);
	if(!foutDem){
		cout << "cutdem.asc file cannot open \n";
		return 1;
	}

	// ArcGIS ASCII Raster header
	foutDem << "ncols "<<Col<<"\n";
	foutDem << "nrows "<<Row<<"\n";
	foutDem << "xllcorner " << SWLON <<"\n";
	foutDem << "yllcorner " << SWLAT <<"\n";
	foutDem << "cellsize " <<Dx << "\n";
	foutDem << "NODATA_value "<< NODATA << "\n";

	for(int i = 0; i < Row; i++){ // write from north
		for(int j = 0;  j< Col; j++){ // write from west
			foutDem << " " << HyshedClipDem[i][j];
		}
		foutDem << "\n";
	}
	cout << "  Writing " << filename << " done." <<endl;
	foutDem.close();

	//sprintf_s(filename, 256, "./output/arcgis/cutDir.asc", projectName);	
	sprintf(filename, "./output/arcgis/cutDir.asc");	
	ofstream foutDir(filename);
	if(!foutDir){
		cout << "cutDir.asc file cannot open \n";
		return 1;
	}
	// ArcGIS ASCII Raster header
	foutDir << "ncols " << Col << "\n";
	foutDir << "nrows " << Row << "\n";
	foutDir << "xllcorner " << SWLON << "\n";
	foutDir << "yllcorner " << SWLAT << "\n";
	foutDir << "cellsize " << Dx << "\n";
	foutDir << "NODATA_value " << NODATA << "\n";

	for(int i = 0; i < Row; i++){ // write from norht
		for(int j = 0; j < Col; j++){ // write from west
			foutDir << " " << HyshedClipDir[i][j];
		}
		foutDir <<"\n";
	}

	cout << "  Writing " << filename << " done." <<endl;
	foutDir.close();
	
	// data processing.
	cout << "3) Processing topography data." << endl;

	//--- Memory allocation ----------------------------------------------------
	class Cell** cellObject;

	cout << "  Memory allocation." << endl;
	cellObject = new class Cell*[Row];
	for (int i=0; i<Row; i++) {
		cellObject[i] = new class Cell[Col];
	}

	// ---Initial setting ------------------------------------------------------
	for (int i=0; i < Row; i++) {
		for(int j=0; j < Col; j++){
			cellObject[i][j].setInit(cellObject, i, j);
		}	
	}

	//--- Setting grid information -----------------------------------------------
	for (int i=0; i<Row; i++) {
		for (int j=0; j<Col; j++) {
			idem = HyshedClipDem[i][j];
			idir = HyshedClipDir[i][j];
			cellObject[i][j].setGridInfo(i, j, idem, idir, SWLAT, SWLON, Row, Col);
		}
	}

	//----------- Setting temporal river mouth -----------------------------------
	for (int i = 1; i < Row-1; i++) {
		for (int j = 1; j < Col-1; j++) {
			cellObject[i][j].setTmpRiverMouth(i, j, Row, Col);
				//  sink ܂͊Cł́AObh̕W -9999 ƂAאڂ
				// 8̃Obh͌Ƃ dir=0 ƂB
		}	
	}

	//----------- Set downstream grid ---------------------------------------------
	cout << "  Setting downstream grid number.\n";
	for (int i=0; i < Row; i++) {
		for (int j=0; j < Col; j++) {
			if (cellObject[i][j].dir <= 0) {
				continue; // skip at the boundary, sea and river mouth grid 
			}
			else {
				cellObject[i][j].setDownStreamGrid(); // setting the downward grid
			}
		}
	}

	//-----------Set upstream grid-----------------------------------------------
	cout << "  Setting upstream grid number.\n";

	for (int i=0; i < Row; i++) {
		for (int j=0; j < Col; j++) {
			if (cellObject[i][j].dir < 0) { // skip at the boundary and sea grid 
				continue;
			}
			if (cellObject[i][j].downStreamGrid == NULL) {
				continue;
			}
			cellObject[i][j].setUpStreamGrid();
		}
	}

	//---------- Set river mouth. ---------------------------------------------
	cout << "  Setting the outlet.\n";
	for (int i=0; i < Row; i++) {
		for (int j=0; j < Col; j++) {
			if (cellObject[i][j].dir == 0) {
				cellObject[i][j].setRiverMouth(); 
			}
		}
	}

	// ----------Calculate the distance from the mouth-------------------------
	cout << "  Calculating the distance from the outlet.\n";

	int distanceNum = 0;
	int icount = 0;

	while (icount < Row*Col) {

		icount = 0;

		for (int i=0; i < Row; i++) {
			for (int j=0; j < Col; j++) {
				if (cellObject[i][j].distanceFromMouth == distanceNum) {
					cellObject[i][j].setDistanceFromMouth();
				}
				else {
					icount++;
				}
			}
		}
		distanceNum++;
	}
	
	//--------------set basin size-------------------------------------------------
	cout << "  Calculating basin area.\n";
	for (int k=distanceNum; k>-1; k--) {
		for (int i=0; i < Row; i++) {
			for (int j=0; j < Col; j++) {
				if (cellObject[i][j].distanceFromMouth == k) {
					cellObject[i][j].setSizeOfBasin(); 
				}
			}
		}
	}

	//---------------riverNumber---------------------------------------------------
	cout << "  Setting river number.\n";

	int rNum;
	rNum = 1;

	for (int i=0; i < Row; i++) {
		for (int j=0; j < Col; j++) {
			if (cellObject[i][j].dir == 0) { // river mouth
				// setting river number at mouth
				cellObject[i][j].riverNumber = rNum;
				rNum++;
			}
		}
	}

	for (int i=0; i < Row; i++) {
		for (int j=0; j < Col; j++) {
			if (cellObject[i][j].sizeOfBasin == 1) { // most upstream grid
				cellObject[i][j].setRiverNumber();
			}
		}
	}


	/*--------------------------------writing output files---------------------------------------------*/
	cout <<"4) Writing output files" << endl;

	FILE *fDem;
	FILE *fDir;
	FILE *fBasin;
	FILE *fRiverNumber;

	//sprintf_s(filename, 256, "./output/modDem.bin", projectName);
	sprintf(filename, "./output/modDem.bin");
	//if(error = fopen_s(&fDem, filename, "wb") != NULL) {
	if ((fDem = fopen(filename, "wb")) == NULL) {
		cout << "modDem.bin file cannot open." << endl;
		exit(1);
	} 

	float elevation;
	for (int i=Row-1; i >= 0; i--) { // write from South to North
		for (int j=0; j<Col; j++) {
			elevation = (float) cellObject[i][j].dem;
			fwrite(&elevation, sizeof(float), 1, fDem);  // write 4 bite real value
		}
	}
	cout << "  Writing " << filename <<endl;
	fclose(fDem);
	
	/*----------------------------modified DEM --------------------------------------------*/
	//sprintf_s(filename, 256, "./output/arcgis/modDem.asc", projectName);
	sprintf(filename, "./output/arcgis/modDem.asc");
	ofstream fout3(filename);
	if (!fout3) {
		cout << "modDem.asc file can not open\n";
		return 1;
	}

	// ArcGIS ASCII Raster header
	fout3 << "ncols " << Col<< "\n";
	fout3 << "nrows " << Row<< "\n";
	fout3 << "xllcorner " << SWLON << "\n";
	fout3 << "yllcorner " << SWLAT << "\n";
	fout3 << "cellsize " << Dx << "\n";
	fout3 << "NODATA_value "<< NODATA;

	for (int i=0; i < Row; i++) {
		for (int j=0; j < Col; j++) {
			if (j == 0) {
				fout3 << "\n";
			}
			fout3 << " " << cellObject[i][j].dem;	
		}
	}
	cout << "  Writing " << filename << " done." <<endl;
	fout3.close();


	/*--------------------------------Distance from the sea --------------------------------------------*/
	/*  This information is not used in the Runoff mode */

	//sprintf_s(filename, 256, "./output/arcgis/distance.asc", projectName);
	sprintf(filename, "./output/arcgis/distance.asc");
	ofstream fout2(filename);
	if (!fout2) {
		cout << "dstance.asc file can not open\n";
		return 1;
	}
	// ArcGIS ASCII Raster header
	fout2 << "ncols " << Col << "\n";
	fout2 << "nrows " << Row << "\n";
	fout2 << "xllcorner " << SWLON << "\n";
	fout2 << "yllcorner " << SWLAT << "\n";
	fout2 << "cellsize " << Dx << "\n";
	fout2 << "NODATA_value " << NODATA;

	for (int i=0; i < Row; i++) {
		for (int j=0; j < Col; j++) {
			if (j == 0) {
				fout2 << "\n";
			}
			fout2 << " " << cellObject[i][j].distanceFromMouth;
		}
	}
	cout << "  Writing " << filename << " done." <<endl;
	fout2.close();


	//--------------------------------Flow direction file-------------------------------------------
	//sprintf_s(filename, 256, "./output/flowDir.bin", projectName);
	sprintf(filename, "./output/flowDir.bin");
	//if(error = fopen_s(&fDir, filename, "wb") != NULL) {
	if((fDir = fopen(filename, "wb")) == NULL) {
		cout << "flowDir.bin file cannot open." << endl;
		exit(1);
	} 

	for (int i=Row-1; i >= 0; i--) { // write from South to North
		for (int j=0; j < Col; j++) {

			if (i==0 || i==Row-1 || j==0 || j==Col-1) { // boundary
				fwrite(&NODATA, sizeof(int), 1, fDir); // write NODATA
			}
			else {
				fwrite(&cellObject[i][j].dir, sizeof(int), 1, fDir); // write 4 bite integer value
			}	
		}
	}
	cout << "  Writing " << filename << " done." <<endl;
	fclose(fDir);

	// ArcGIS ASCII Raster header 
	//sprintf_s(filename, 256, "./output/arcgis/flowDir.asc", projectName);
	sprintf(filename, "./output/arcgis/flowDir.asc");
	ofstream fout(filename);
	if (!fout) {
		cout << "file can not open\n";
		return 1;
	}
	// ArcGIS ASCII Raster header
	fout << "ncols " << Col << "\n";
	fout << "nrows " << Row << "\n";
	fout << "xllcorner " << SWLON << "\n";
	fout << "yllcorner " << SWLAT << "\n";
	fout << "cellsize " << Dx << "\n";
	fout << "NODATA_value "<< NODATA;  // DNODATA = -9999

	for (int i=0; i < Row; i++) {
		for (int j=0; j < Col; j++) {
			if (j == 0) {
				fout << "\n";
			}

			// change the Flow direction to ESRI arc grid format
			if (cellObject[i][j].dir ==3){
				cellObject[i][j].dir = 1;
			}
			else if (cellObject[i][j].dir ==4){
				cellObject[i][j].dir=2;
			}		
			else if (cellObject[i][j].dir ==5){
				cellObject[i][j].dir=4;
			}		
			else if (cellObject[i][j].dir ==6){
				cellObject[i][j].dir=8;
			}	
			else if (cellObject[i][j].dir ==7){
				cellObject[i][j].dir=16;
			}	
			else if (cellObject[i][j].dir ==8){
				cellObject[i][j].dir=32;
			}	
			else if (cellObject[i][j].dir ==1){
				cellObject[i][j].dir=64;
			}	
			else if (cellObject[i][j].dir ==2){
				cellObject[i][j].dir=128;
			}
			else if (cellObject[i][j].dir ==0){
				cellObject[i][j].dir=0;
			}				
			else if (cellObject[i][j].dir < 0) { 
				cellObject[i][j].dir = NODATA;  
			}				
			else {
				cout << i <<"-"<< j << "-"<< cellObject[i][j].dir << "cellObject.down is wrong\n";
			}
			fout << " " << cellObject[i][j].dir;
		}
	}
	cout << "  Writing " << filename << " done." <<endl;
	fout.close();
	

	//-------------------------Flow accumulation file--------------------------------------------
	//sprintf_s(filename, 256, "./output/flowAcc.bin", projectName);
	sprintf(filename, "./output/flowAcc.bin");
	//if(error = fopen_s(&fBasin, filename, "wb") != NULL) {
	if((fBasin = fopen(filename, "wb")) == NULL) {
		cout << "flowAcc.bin file cannot open." << endl;
		exit(1);
	} 

	for (int i=Row-1; i >= 0; i--) { // write from South to North
		for (int j=0; j < Col; j++) {
			
			if (i==0 || i==Row-1 || j==0 || j==Col-1) { // boundary
				fwrite(&NODATA, sizeof(int), 1, fBasin); // write NODATA
			}
			else {
				fwrite(&cellObject[i][j].sizeOfBasin, sizeof(int), 1, fBasin); // write 4 bite integer value
			}
		}
	}
	cout << "  Writing " << filename << " done." <<endl;
	fclose(fBasin);

	// ArcGIS 
	//sprintf_s(filename, 256, "./output/arcgis/flowAcc.asc", projectName);
	sprintf(filename, "./output/arcgis/flowAcc.asc");
	ofstream fout1(filename);
	if(!fout1){
		cout << "flowAcc.asc file can not open\n";
		return 1;
	}
	fout1 << "ncols " << Col << "\n";
	fout1 << "nrows " << Row << "\n";
	fout1 << "xllcorner " << SWLON << "\n";
	fout1 << "yllcorner " << SWLAT << "\n";
	fout1 << "cellsize " << Dx << "\n";
	fout1 << "NODATA_value " << NODATA;

	for (int i=0; i < Row; i++) {
		for (int j=0; j < Col; j++) {
			if(j == 0) {
				fout1 << "\n";
			}
			if (i==0 || i==Row-1 || j==0 || j==Col-1) {
				fout1 << " " << NODATA;
			}
			else {
				fout1 << " " << cellObject[i][j].sizeOfBasin;
			}
		}
	}
	cout << "  Writing " << filename << " done." <<endl;
	fout1.close();

	/*--------------------------------riverNumber---------------------------------------------*/
	//sprintf_s(filename, 256, "./output/riverNum.bin", projectName);
	sprintf(filename, "./output/riverNum.bin");
	//if(error = fopen_s(&fRiverNumber, filename, "wb") != NULL) {
	if((fRiverNumber = fopen(filename, "wb")) == NULL) {
		cout << "riverNum.bin file cannot open." << endl;
		exit(1);
	} 

	for (int i=Row-1; i >= 0; i--) { // write from South to North
		for (int j=0; j < Col; j++) {
			if (i==0 || i==Row-1 || j==0 || j==Col-1) {
				fwrite(&NODATA, sizeof(int), 1, fRiverNumber); // write NODATA
			}
			else {
				fwrite(&cellObject[i][j].riverNumber, sizeof(int), 1, fRiverNumber); // write 4 bite integer value
			}
		}
	}
	cout << "  Writing " << filename << " done." <<endl;
	fclose(fRiverNumber);

	// ArcGIS
	//sprintf_s(filename, 256, "./output/arcgis/riverNum.asc", projectName);
	sprintf(filename, "./output/arcgis/riverNum.asc");
	ofstream fout4(filename);
	if (!fout4) {
		cout << "ESRI RiverNum.asc riverNumber file can not open\n";
		return 1;
	}
	fout4 << "ncols " << Col << "\n";
	fout4 << "nrows " << Row << "\n";
	fout4 << "xllcorner " << SWLON << "\n";
	fout4 << "yllcorner " << SWLAT << "\n";
	fout4 << "cellsize " << Dx << "\n";
	fout4 << "NODATA_value "<< NODATA;
	
	for (int i=0; i < Row; i++) {
		for (int j=0; j < Col; j++) {
			if(j == 0) {
				fout4 << "\n";
			}
			if (i==0 || i==Row-1 || j==0 || j==Col-1) {
				fout4 << " " << NODATA;
			}
			else {
				fout4 << " " << cellObject[i][j].riverNumber;
			}
		}
	}
	cout << "  Writing " << filename << " done." <<endl;
	fout4.close();

	//--- riverNumber list ŉƂȂObh̏o
	//sprintf_s(filename, 256, "./output/riverNumList.dat", projectName);
	sprintf(filename, "./output/riverNumList.dat");
	ofstream fRiverNumList(filename);
	if(!fRiverNumList){
		cout << "riverNumList.dat file can not open\n";
		return 1;
	}

	for (int i=0; i < Row; i++) { // row number
		for (int j=0; j < Col; j++) { // column number 
			if (cellObject[i][j].dir == 0) {  // river mouth
				fRiverNumList << cellObject[i][j].riverNumber 
					<< " " << cellObject[i][j].sizeOfBasin // area
					<< " " << j+1  // column number from West
					<< " " << (Row-1)-i+1  // row number from South
					<< " " << cellObject[i][j].longitude // longitude (second)
					<< " " << cellObject[i][j].latitude // latitude (second)
					<< " " << double (cellObject[i][j].longitude)/3600.0 // longitude (degree)
					<< " " << double (cellObject[i][j].latitude)/3600.0 // latitude (degee)
					<< endl;
			}
		}
	}
	cout << "  Writing " << filename << " done." <<endl;
	fRiverNumList.close();

	delete[] cellObject;

	cout << endl;
	cout << "Program ends!" << endl;
	// cout << "Program ends. Hit any key!" << endl;
	// char ch;
	// cin.get(ch);
	
	return 0;
}

//========== member functions definition ==================

/*  Initialize flow direction, adjesant grid number and lower grid number */
void Cell::setInit(class Cell** cell2, int ii, int jj)
{
	// set initial values for data members
	distanceFromMouth = -9999;
	sizeOfBasin = 1;
	riverNumber = -9999;

	downStreamGrid=NULL;

	for (int i=0; i < 8; i++) {
		neighborGrid[i] = NULL;
	}

	for (int i=0; i < 7; i++) {
		upStreamGrid[i] = NULL;
	}

	// set address of neighbor grid cell
	neighborGrid[0]=&cell2[ii-1][jj];
	neighborGrid[1]=&cell2[ii-1][jj+1];
	neighborGrid[2]=&cell2[ii][jj+1];
	neighborGrid[3]=&cell2[ii+1][jj+1];
	neighborGrid[4]=&cell2[ii+1][jj];
	neighborGrid[5]=&cell2[ii+1][jj-1];
	neighborGrid[6]=&cell2[ii][jj-1];
	neighborGrid[7]=&cell2[ii-1][jj-1];
}

void Cell::setGridInfo (
	int irow, // row counter
	int icol, // column counter
	int dem1, // elevation at grid center
	int dir1,  // flow direction fot the grid
	double SWlat, // South west corner of latitude
	double SWlon, // South west corner of longitude
	int Row, // number of row for the study area
	int Col  //number of column for the study area
)
{
	latitude  = (int)(SWlat*3600) + 30*(Row-irow-1) + 15;
	longitude = (int)(SWlon*3600) + 30*icol + 15;

	longitudinalLength = 926.6*cos(latitude/3600.0/180.0*3.14);

	dem = dem1;

	// HydroSHED flow direction sign convention is diffrent from the convention used 
	// with 1K-FRM. Cnvert Hydroshed direction.

	if (dir1 == 1) { dir = 3;} // EE: 3
	else if (dir1 == 2) { dir = 4;} // SE: 4
	else if (dir1 == 4) { dir = 5;} // SS: 5
	else if (dir1 == 8) { dir = 6;} // SW: 6
	else if (dir1 == 16) { dir = 7;} // WW: 7
	else if (dir1 == 32) { dir = 8;} // NW: 8
	else if (dir1 == 64) { dir = 1;} // NN: 1
	else if (dir1 == 128) { dir = 2;} // NE: 2
	else if (dir1 == 0) { dir = 0;} // river mouth
	else if (dir1 == -1) { // inland sink or pit
		// According to HydroShed, -1 is a inland sink or pit. 
		// The cell is set as sea or pit.
		dir = -1;
		sizeOfBasin = -9999;
	} 
	else if (dir1 == -9999) { // nodata
		dir = -9999;
		sizeOfBasin = -9999;
	} 
	else {
		cout << "setElabation: Flow direction is wrong." << endl;
		cout << "row = " << irow << " col = " << icol << " dir = " << dir1 << endl;
		exit(1);
	}

	// commented out on Oct. 27, 2012 by Tachikawa
	//if (dem1  < 0.0){ // If elevation < 0, set no data.
	//	dir = -1;
	//}

	// set boundary of study area
	if (irow==0 || icol==0 || irow== (Row-1) || icol==(Col-1)) { 
		// set boundary of study area
		dir = -9; // set boundary grid direction as -9.
		dem = -9999; // set boundary elevation as -sink.
	}
}


// Setting temporal river mouth 
void Cell::setTmpRiverMouth(int irow, int icol, int Row, int Col)
{
	if (irow==1 || icol==1 || irow== (Row-2) || icol==(Col-2)) { 
		// set one line inside of study area boundary as river mouth 
		// Ȃ΁Ä拫Ê͌Ƃ
		if (dem != -9999) dir = 0; // river mouth: 0
	}

	if (dir == -1) { // if this grid is a grid cell of inside pit or sea
		dem = -9999; // set the elevation as sea.
		for(int i=0; i < 8; i++) {
			neighborGrid[i]->dir = 0; // set the surrounding grids as river mouth.
		}
	}
}

// Setting river mouth from temporally defined mouthes.
void Cell::setRiverMouth()
{
	int mouthOrNot = 0;
	for (int i = 0; i < 7; i++) {
		if (upStreamGrid[i] != NULL) {
			mouthOrNot = 1; // this grid is river mouth.
		}
	}

	if (mouthOrNot == 1) {
		distanceFromMouth = 0;
	}
	else { // this grid is not river mouth.
		dir = -9999;
		dem = -9999;
		sizeOfBasin = -9999;
	} 
}

// setting diwnstram grid
void Cell::setDownStreamGrid()
{
	if (1 <= dir && dir <= 8) {
		downStreamGrid = neighborGrid[dir-1];
	}
	else {
		downStreamGrid = NULL;
	}
}

// setting upstream grid
void Cell::setUpStreamGrid()
{
	for (int i=0; i < 7; i++) {
		if (downStreamGrid->upStreamGrid[i] == NULL) { 
			downStreamGrid->upStreamGrid[i] = this;
			break;
		}
	}
}

// calculate the distance from the river mouth 
void Cell::setDistanceFromMouth()
{
	for (int i=0; i < 7; i++) {
		if (upStreamGrid[i] != NULL) {
			upStreamGrid[i]->distanceFromMouth = distanceFromMouth + 1;
		}
	}				
}

// calculate the number of grids from upstream 
void Cell::setSizeOfBasin()
{	
	for (int i=0; i<7; i++) {
		if (upStreamGrid[i] != NULL) {
			sizeOfBasin += upStreamGrid[i]->sizeOfBasin;
			// initially sizeOfBasin is set to 1 see setInit function
		}
	}
}

// set the river number 
void Cell::setRiverNumber()
{
	int rNum;
	class Cell* object;
	
	// repeat until this reaches to the outlet and get the river number
	object = this;
	while (object->riverNumber == -9999) {
		if (object->downStreamGrid == NULL) {
			break;
		}
		object = object->downStreamGrid;
	}
	rNum = object->riverNumber;

	// again repeat until this reaches to the outlet and get the 
	// river number. Assign the river number to the grids on the 
	// pathway.
	object = this;
	while (object->riverNumber == -9999) {
		if (object->downStreamGrid == NULL) {
			break;
		}
		object->riverNumber = rNum;
		object = object->downStreamGrid;
	}
}


